/* symtab.c   The GPLBASIC interpreter
   Copyright (C) 1997 L.C. Benschop, Eindhoven, the Netherlands
   */

#include "gplbasic.h"
#include "symtab.h"
#include "token.h"
#include "errors.h"
#include "strings.h"
#include "basicio.h"
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <time.h>

VARSYM * hashed_syms[S_HASHENTRIES];
LABSYM * hashed_labs[S_HASHENTRIES];

int symareas;
VARSYM * symptr;
int syms_used;  /* area to store symbol table entries */
int labareas;
LABSYM * labptr;
int labs_used;  /* area to store label entries */


char *labname(LABSYM *);

void clear_symtabs(void)
     /* the memory areas themselves are cleared by clear_prog() */
{
  int i;
  for(i=0;i<S_HASHENTRIES;i++) {
    hashed_syms[i]=0;
    hashed_labs[i]=0;
  }
  symptr=0;
  labptr=0;
}

static unsigned char * alloc_name (void) 
     /* Allocate memory for the symbol name */
{
  int l;
  unsigned char * np;
  l=strlen(tokname)+1;
  if ( chars_used + l > S_STRCONST || charconst == 0) {
    if(memareas==N_MEMAREAS) doerror(GE_MEM,0,0);
    if( !(mem_areas[memareas]=malloc(S_STRCONST))) doerror(GE_MEM,0,0);
    charconst=mem_areas[memareas++];
    charareas++;
    chars_used=0;
  }
  np=charconst;
  strcpy(charconst,tokname);
  charconst+=l;
  chars_used+=l;
  return np;
}

static LABSYM * alloc_lab()
     /* Allocate new label entry */
{
 LABSYM *p;
 if(labs_used==S_SYMCHUNK ||labptr==0) { /* allocate new labeltab memory */
   if(memareas==N_MEMAREAS) doerror(GE_MEM,0,0);
   if( !(mem_areas[memareas]=malloc(S_SYMCHUNK * sizeof(LABSYM)))) 
     doerror(GE_MEM,0,0);
   labptr=mem_areas[memareas++];
   labareas++;
   labs_used=0;
 }
 p=labptr++;
 labs_used++;
 return p; 
}

LABSYM * look_label(void)
{
 unsigned hash;
 unsigned long lnum;
 LABSYM * p;
 unsigned char * np;
 if(isdigit(tokname[0])) { /* numeric label */
   lnum=atol(tokname);
   hash=(lnum % S_HASHENTRIES);
   p=hashed_labs[hash];
   while(p) {
    if(p->id.num==lnum && (p->flags&1)==0) break;
    else p=p->next;
   }
   if(!p) {
     p=alloc_lab();
     p->next=hashed_labs[hash];
     hashed_labs[hash]=p;
     p->flags=0;
     p->id.num=lnum;
     p->contents=0;
   }
 } else { /* alphanumeric label */
   hash=tokhash % S_HASHENTRIES;
   p=hashed_labs[hash];
   while(p) {
     if( (!strcmp(p->id.name,tokname)) && (p->flags&1) ) break;
     else p=p->next;
   }
   if(!p) {
     p=alloc_lab();
     np=alloc_name();
     p->next=hashed_labs[hash];
     hashed_labs[hash]=p;
     p->id.name=np;
     p->contents=0;
     p->flags=1;
   }
 }
 return p;
}

VARSYM * look_var(void)
{
 unsigned hash;
 VARSYM * p;
 unsigned char * np;
 hash=(tokhash ^ ((vartype&127)<<8)) % S_HASHENTRIES;
 p=hashed_syms[hash];
 while(p) {
   if((!strcmp(p->name,tokname)) && (p->type&127) == (vartype&127) ) {
     break; /* found matching name */
   } else {
     p=p->next;
   }
 }
 if(!p) { /* symbol not found, so create it */
   np=alloc_name();
   if(syms_used==S_SYMCHUNK ||symptr==0) { /* allocate new symtab memory */
     if(memareas==N_MEMAREAS) doerror(GE_MEM,0,0);
     if( !(mem_areas[memareas]=malloc(S_SYMCHUNK * sizeof(VARSYM)))) 
       doerror(GE_MEM,0,0);
     symptr=mem_areas[memareas++];
     symareas++;
     syms_used=0;
   }
   p=symptr++;
   syms_used++;
   p->next=hashed_syms[hash];
   hashed_syms[hash]=p;
   p->name=np;
   p->contents.l=0;
   p->type=0; /* set type to 0 to indicate new symbol */     
 }
 return p;
}

void check_unresolved(void)
{
 int i;
 LABSYM *p;VARSYM *vp;
 for(i=0;i<S_HASHENTRIES;i++) {
   p=hashed_labs[i];
   while(p) {
     if( !(p->flags & 2) ) doerror(GE_LAB,labname(p),0);
     p=p->next;
   }
   vp=hashed_syms[i];
   while(vp) {
     if( (vp->type & TY_FUNC) && 
     !( ((CELLTYPE*)(vp->contents.p))->l & 256) ) 
                doerror(GE_SBR,vp->name,0);
     vp=vp->next;
   }
 }
}

void clear_vars(void)
{
  int i;
  CELLTYPE *p1;
  VARSYM *p;
  unsigned char *ps,*qs;
  stkptr = main_stack + S_MAINSTACK - 1;
  stkptr->l = KW_END; /* push end marker onto stack */
  fstkptr = float_stack + S_FLTSTACK;
  strstkptr = string_stack;
  strstkmark = string_stack;
  frameptr=stkptr;
  flframe=fstkptr;
  for(i=0;i<S_HASHENTRIES;i++) {
    p=hashed_syms[i];
    while(p) {
      if( p->type == TY_INT||p->type ==TY_LONG) p->contents.l=0;
      else if( p->type == TY_STR) p->contents.p=short_strings;
      else if( p->type == TY_DOUBLE||p->type==TY_SINGLE) p->contents.d=0;
      else if( (p->type & 192)==64 ) { /* is it an array?*/
	if( p->contents.p) { /* array is dimensioned? */
	  p1=p->contents.p;
	  if( p1[1].p) free( p1[1].p ); /* free array contents */
	  free(p1); /* free array descriptor */
	  p->contents.p=0;
	}
      }
      p=p->next;
    }
  }
  ps=stringareas;
  while(ps) {
    qs=*(char **)ps;
    free(ps);
    ps=qs;
  }
  stringareas=0;
  cur_data=datastart;
  data_used=0;
  srand(time(0));
  init_io();
}

char * labname(LABSYM *p)
{
  static char numbuf[21];
  if(p->flags & 1) {
   return p->id.name;
  } else {
    sprintf(numbuf,"%d",p->id.num);
    return numbuf;
  }  
}

void list_vars(void)
{
  int i,j,a;
  VARSYM *p;
  CELLTYPE *p1;
  for(i=0;i<S_HASHENTRIES;i++) {
    p=hashed_syms[i];j=0;
    while(p) {
      printf("%3d %s",j,p->name);
      switch((p->type)&7) {
      case TY_INT: printf("%");break;
      case TY_LONG: printf("&");break;
      case TY_SINGLE: printf("!");break;
      case TY_DOUBLE: printf("#");break;
      case TY_STR: printf("$");break;
      }
      if(p->type & 128) {
	printf(" <function>");
      } else if(p->type & 64) { /*array*/
	printf("(");
	if(p->contents.p) {
	  p1=p->contents.p;
	  for(a=0;a<p1[0].l;a++) {
	    if(a)printf(",");
	    printf("%d",p1[a+2].l);
	  }
	}
	printf(")");
      } else { /* normal variable */
	printf("=");
	a=p->type&7;
	if(a==TY_INT || a==TY_LONG) print_int(p->contents.l);
	else if(a==TY_DOUBLE ||a==TY_SINGLE) print_float(p->contents.d);
	else {
	  printf("\"");
	  print_str(p->contents.p);
	  printf("\"");
	}
      }
      printf("\n");
      p=p->next;
      j++;
    }
  }
}
